import {fireEvent, render} from '@testing-library/react'
import React from 'react'
import {ThemeProvider} from '../ThemeProvider'
import {SubTreeState, TreeView} from './TreeView'

// TODO: Move this function into a shared location
function renderWithTheme(
  ui: Parameters<typeof render>[0],
  options?: Parameters<typeof render>[1]
): ReturnType<typeof render> {
  return render(<ThemeProvider>{ui}</ThemeProvider>, options)
}

describe('Markup', () => {
  it('uses tree role', () => {
    const {queryByRole} = renderWithTheme(
      <TreeView aria-label="Test tree">
        <TreeView.Item>Item 1</TreeView.Item>
        <TreeView.Item>Item 2</TreeView.Item>
        <TreeView.Item>Item 3</TreeView.Item>
      </TreeView>
    )

    const root = queryByRole('tree')

    expect(root).toHaveAccessibleName('Test tree')
  })

  it('uses treeitem role', () => {
    const {queryAllByRole} = renderWithTheme(
      <TreeView aria-label="Test tree">
        <TreeView.Item>Item 1</TreeView.Item>
        <TreeView.Item>Item 2</TreeView.Item>
        <TreeView.Item>Item 3</TreeView.Item>
      </TreeView>
    )

    const items = queryAllByRole('treeitem')

    expect(items).toHaveLength(3)
  })

  it('hides subtrees by default', () => {
    const {queryByRole} = renderWithTheme(
      <TreeView aria-label="Test tree">
        <TreeView.Item>
          Parent
          <TreeView.SubTree>
            <TreeView.Item>Child</TreeView.Item>
          </TreeView.SubTree>
        </TreeView.Item>
      </TreeView>
    )

    const parentItem = queryByRole('treeitem', {name: 'Parent'})
    const subtree = queryByRole('group')

    expect(parentItem).toHaveAttribute('aria-expanded', 'false')
    expect(subtree).toBeNull()
  })

  it('uses aria-current', () => {
    const {getByRole} = renderWithTheme(
      <TreeView aria-label="Test tree">
        <TreeView.Item>Item 1</TreeView.Item>
        <TreeView.Item current>Item 2</TreeView.Item>
        <TreeView.Item>Item 3</TreeView.Item>
      </TreeView>
    )

    const currentItem = getByRole('treeitem', {name: 'Item 2'})

    expect(currentItem).toHaveAttribute('aria-current', 'true')
  })

  it('expands the path to the current item (level 2) by default', () => {
    const {getByRole} = renderWithTheme(
      <TreeView aria-label="Test tree">
        <TreeView.Item>
          Item 1
          <TreeView.SubTree>
            <TreeView.Item>Item 1.1</TreeView.Item>
          </TreeView.SubTree>
        </TreeView.Item>
        <TreeView.Item>
          Item 2
          <TreeView.SubTree>
            <TreeView.Item>Item 2.1</TreeView.Item>
            <TreeView.Item current>
              Item 2.2
              <TreeView.SubTree>
                <TreeView.Item>Item 2.2.1</TreeView.Item>
              </TreeView.SubTree>
            </TreeView.Item>
          </TreeView.SubTree>
        </TreeView.Item>
        <TreeView.Item>Item 3</TreeView.Item>
      </TreeView>
    )

    const item1 = getByRole('treeitem', {name: 'Item 1'})
    const item2 = getByRole('treeitem', {name: 'Item 2'})
    const item22 = getByRole('treeitem', {name: 'Item 2.2'})
    const item221 = getByRole('treeitem', {name: 'Item 2.2.1'})

    // Item 1 should not be expanded because it is not the parent of the current item
    expect(item1).toHaveAttribute('aria-expanded', 'false')

    // Item 2 should be expanded because it is the parent of the current item
    expect(item2).toHaveAttribute('aria-expanded', 'true')

    // Item 2.2 should be expanded because it is the current item
    expect(item22).toHaveAttribute('aria-expanded', 'true')

    // Item 2.2 should have an aria-current value of true
    expect(item22).toHaveAttribute('aria-current', 'true')

    // Item 2.2.1 should be visible because it is a child of the current item
    expect(item221).toBeVisible()
  })

  it('expands the path to the current item (level 3) by default', () => {
    const {getByRole} = renderWithTheme(
      <TreeView aria-label="Test tree">
        <TreeView.Item>
          Item 1
          <TreeView.SubTree>
            <TreeView.Item>Item 1.1</TreeView.Item>
          </TreeView.SubTree>
        </TreeView.Item>
        <TreeView.Item>
          Item 2
          <TreeView.SubTree>
            <TreeView.Item>Item 2.1</TreeView.Item>
            <TreeView.Item>
              Item 2.2
              <TreeView.SubTree>
                <TreeView.Item current>Item 2.2.1</TreeView.Item>
              </TreeView.SubTree>
            </TreeView.Item>
          </TreeView.SubTree>
        </TreeView.Item>
        <TreeView.Item>Item 3</TreeView.Item>
      </TreeView>
    )

    const item1 = getByRole('treeitem', {name: 'Item 1'})
    const item2 = getByRole('treeitem', {name: 'Item 2'})
    const item22 = getByRole('treeitem', {name: 'Item 2.2'})
    const item221 = getByRole('treeitem', {name: 'Item 2.2.1'})

    // Item 1 should not be expanded because it is not the parent of the current item
    expect(item1).toHaveAttribute('aria-expanded', 'false')

    // Item 2 should be expanded because it is the parent of the current item
    expect(item2).toHaveAttribute('aria-expanded', 'true')

    // Item 2.2 should be expanded because it is the current item
    expect(item22).toHaveAttribute('aria-expanded', 'true')

    // Item 2.2.1 should be the current item
    expect(item221).toHaveAttribute('aria-current', 'true')
  })

  it('expands the path to the current item when the current item is changed', () => {
    function TestTree() {
      const [current, setCurrent] = React.useState('item1')
      return (
        <div>
          <button onClick={() => setCurrent('item2')}>Jump to Item 2</button>
          <TreeView aria-label="Test tree">
            <TreeView.Item current={current === 'item1'}>Item 1</TreeView.Item>
            <TreeView.Item current={current === 'item2'}>
              Item 2
              <TreeView.SubTree>
                <TreeView.Item current={current === 'item2.1'}>Item 2.1</TreeView.Item>
              </TreeView.SubTree>
            </TreeView.Item>
            <TreeView.Item current={current === 'item3'}>Item 3</TreeView.Item>
          </TreeView>
        </div>
      )
    }

    const {getByRole, getByText} = renderWithTheme(<TestTree />)

    const item1 = getByRole('treeitem', {name: 'Item 1'})
    const item2 = getByRole('treeitem', {name: 'Item 2'})

    // Item 1 should have an aria-current value of true
    expect(item1).toHaveAttribute('aria-current', 'true')

    // Item 2 should not be expanded because it is not the current item or the parent of the current item
    expect(item2).toHaveAttribute('aria-expanded', 'false')

    // Click the button to change the current item to Item 2
    fireEvent.click(getByText('Jump to Item 2'))

    // Item 1 should not have an aria-current value
    expect(item1).not.toHaveAttribute('aria-current')

    // Item 2 should be expanded because it is the current item
    expect(item2).toHaveAttribute('aria-expanded', 'true')

    // Item 2.1 should be visible because it is a child of the current item
    expect(getByRole('treeitem', {name: 'Item 2.1'})).toBeVisible()
  })

  it('should be described by leading visuals', () => {
    const {getByLabelText} = renderWithTheme(
      <TreeView aria-label="Test tree">
        <TreeView.Item>
          <TreeView.LeadingVisual label="leading">
            <svg aria-hidden={true} />
          </TreeView.LeadingVisual>
          Item 1
        </TreeView.Item>
        <TreeView.Item>
          <TreeView.LeadingVisual>
            <svg aria-hidden={true} />
          </TreeView.LeadingVisual>
          Item 2
        </TreeView.Item>
      </TreeView>
    )
    const item = getByLabelText(/Item 1/)
    expect(item).toHaveAccessibleDescription('leading')

    const noDescription = getByLabelText(/Item 2/)
    expect(noDescription).not.toHaveAccessibleDescription()
  })

  it('should be described by trailing visuals', () => {
    const {getByLabelText} = renderWithTheme(
      <TreeView aria-label="Test tree">
        <TreeView.Item>
          Item 1
          <TreeView.TrailingVisual label="trailing">
            <svg aria-hidden={true} />
          </TreeView.TrailingVisual>
        </TreeView.Item>
        <TreeView.Item>
          Item 2
          <TreeView.TrailingVisual>
            <svg aria-hidden={true} />
          </TreeView.TrailingVisual>
        </TreeView.Item>
      </TreeView>
    )
    const item = getByLabelText(/Item 1/)
    expect(item).toHaveAccessibleDescription('trailing')

    const noDescription = getByLabelText(/Item 2/)
    expect(noDescription).not.toHaveAccessibleDescription()
  })

  it('should be described by leading and trailing visuals', () => {
    const {getByLabelText} = renderWithTheme(
      <TreeView aria-label="Test tree">
        <TreeView.Item>
          <TreeView.LeadingVisual label="leading">
            <svg aria-hidden={true} />
          </TreeView.LeadingVisual>
          Item 1
          <TreeView.TrailingVisual label="trailing">
            <svg aria-hidden={true} />
          </TreeView.TrailingVisual>
        </TreeView.Item>
        <TreeView.Item>
          <TreeView.LeadingVisual>
            <svg aria-hidden={true} />
          </TreeView.LeadingVisual>
          Item 2
          <TreeView.TrailingVisual>
            <svg aria-hidden={true} />
          </TreeView.TrailingVisual>
        </TreeView.Item>
      </TreeView>
    )
    const item = getByLabelText(/Item 1/)
    expect(item).toHaveAccessibleDescription('leading trailing')

    const noDescription = getByLabelText(/Item 2/)
    // Note: it seems the computed description here is a string with a single
    // space due to the implementation of `aria-describedby`. We currently set
    // both trailing and visual and when the nodes are not found in
    // `aria-describedby="uuid-leading uuid-trailing"` then it computes to a
    // space
    expect(noDescription).toHaveAccessibleDescription(' ')
  })
})

describe('Keyboard interactions', () => {
  describe('ArrowDown', () => {
    it('moves focus to the next visible treeitem', () => {
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item defaultExpanded>
            Item 1
            <TreeView.SubTree>
              <TreeView.Item>Item 1.1</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
          <TreeView.Item>
            Item 2
            <TreeView.SubTree>
              <TreeView.Item>Item 2.1</TreeView.Item>
              <TreeView.Item>Item 2.2</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
          <TreeView.Item>Item 3</TreeView.Item>
        </TreeView>
      )

      const item1 = getByRole('treeitem', {name: 'Item 1'})
      const item11 = getByRole('treeitem', {name: 'Item 1.1'})
      const item2 = getByRole('treeitem', {name: 'Item 2'})
      const item3 = getByRole('treeitem', {name: 'Item 3'})

      // Focus first item
      item1.focus()

      // Press ↓
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})

      // item 1.1 should be focused
      expect(item11).toHaveFocus()

      // Press ↓
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})

      // item 2 should be focused
      expect(item2).toHaveFocus()

      // Press ↓
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})

      // item 3 should have focus (skips item 2.1 and item 2.2 because they are hidden)
      expect(item3).toHaveFocus()

      // Press ↓
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})

      // focus should not change (item 3 is the last visible treeitem)
      expect(item3).toHaveFocus()
    })
  })

  describe('ArrowUp', () => {
    it('moves focus to the previous visible treeitem', () => {
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item defaultExpanded>
            Item 1
            <TreeView.SubTree>
              <TreeView.Item>Item 1.1</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
          <TreeView.Item>
            Item 2
            <TreeView.SubTree>
              <TreeView.Item>Item 2.1</TreeView.Item>
              <TreeView.Item>Item 2.2</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
          <TreeView.Item>Item 3</TreeView.Item>
        </TreeView>
      )

      const item1 = getByRole('treeitem', {name: 'Item 1'})
      const item11 = getByRole('treeitem', {name: 'Item 1.1'})
      const item2 = getByRole('treeitem', {name: 'Item 2'})
      const item3 = getByRole('treeitem', {name: 'Item 3'})

      // Focus first item
      item1.focus()

      // Press ↓ 4 times to move aria-activedescendant to item 3
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})

      // item 3 should be focused
      expect(item3).toHaveFocus()

      // Press ↑
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowUp'})

      // Item 2 should have focus (skips item 2.1 and item 2.2 because they are hidden)
      expect(item2).toHaveFocus()

      // Press ↑
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowUp'})

      // Item 1.1 should be focused
      expect(item11).toHaveFocus()

      // Press ↑
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowUp'})

      // Item 1 should be focused
      expect(item1).toHaveFocus()

      // Press ↑
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowUp'})

      // Focus should not change (item 1 is the first visible treeitem)
      expect(item1).toHaveFocus()
    })
  })

  describe('ArrowLeft', () => {
    it('collapses an expanded item', () => {
      const {getByRole, queryByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item defaultExpanded>
            Parent
            <TreeView.SubTree>
              <TreeView.Item>Child</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
        </TreeView>
      )

      const parentItem = getByRole('treeitem', {name: 'Parent'})
      let subtree = queryByRole('group')

      // aria-expanded should be true
      expect(parentItem).toHaveAttribute('aria-expanded', 'true')

      // Subtree should be visible
      expect(subtree).toBeVisible()

      // Focus first item
      parentItem.focus()

      // Press ←
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowLeft'})

      // aria-expanded should now be false
      expect(parentItem).toHaveAttribute('aria-expanded', 'false')

      // Parent item should still be focused
      expect(parentItem).toHaveFocus()

      subtree = queryByRole('group')

      // Subtree should now be hidden
      expect(subtree).toBeNull()
    })

    it('does nothing on a root-level collapsed item', () => {
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item>
            Parent
            <TreeView.SubTree>
              <TreeView.Item>Child</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
        </TreeView>
      )

      const parentItem = getByRole('treeitem', {name: 'Parent'})

      // aria-expanded should be false by default
      expect(parentItem).toHaveAttribute('aria-expanded', 'false')

      // Focus first item
      parentItem.focus()

      // Press ←
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowLeft'})

      // aria-expanded should still be false
      expect(parentItem).toHaveAttribute('aria-expanded', 'false')

      // Focus should not change
      expect(parentItem).toHaveFocus()
    })

    it('does nothing on a root-level end item', () => {
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item>Item</TreeView.Item>
        </TreeView>
      )

      const item = getByRole('treeitem', {name: 'Item'})

      // Focus first item
      item.focus()

      // Press ←
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowLeft'})

      // Focus should not change
      expect(item).toHaveFocus()
    })

    it('moves focus to parent of end item', () => {
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item defaultExpanded>
            Parent
            <TreeView.SubTree>
              <TreeView.Item>Child 1</TreeView.Item>
              <TreeView.Item>Child 2</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
        </TreeView>
      )

      const parentItem = getByRole('treeitem', {name: 'Parent'})
      const child2 = getByRole('treeitem', {name: 'Child 2'})

      // Focus fist item
      parentItem.focus()

      // Press ↓ 2 times to move focus to child 2
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})

      // Child 2 should be focused
      expect(child2).toHaveFocus()

      // Press ←
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowLeft'})

      // Parent item should be focused
      expect(parentItem).toHaveFocus()
    })

    it('moves focus to parent of collapsed item', () => {
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item defaultExpanded>
            Parent
            <TreeView.SubTree>
              <TreeView.Item>Child</TreeView.Item>
              <TreeView.Item>
                Nested parent
                <TreeView.SubTree>
                  <TreeView.Item>Nested child</TreeView.Item>
                </TreeView.SubTree>
              </TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
        </TreeView>
      )

      const parentItem = getByRole('treeitem', {name: 'Parent'})
      const nestedParentItem = getByRole('treeitem', {name: 'Nested parent'})

      // Focus first item
      parentItem.focus()

      // Press ↓ 2 times to move focus to nested parent
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})

      // Nested parent item should be focused
      expect(nestedParentItem).toHaveFocus()

      // Press ←
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowLeft'})

      // Parent item should be focused
      expect(parentItem).toHaveFocus()
    })
  })

  describe('ArrowRight', () => {
    it('expands a collapsed item', () => {
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item>
            Parent
            <TreeView.SubTree>
              <TreeView.Item>Child</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
        </TreeView>
      )

      const parentItem = getByRole('treeitem', {name: 'Parent'})

      // aria-expanded should be false by default
      expect(parentItem).toHaveAttribute('aria-expanded', 'false')

      // Focus first item
      parentItem.focus()

      // Press →
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowRight'})

      // aria-expanded should now be true
      expect(parentItem).toHaveAttribute('aria-expanded', 'true')

      // Parent item should still be focused
      expect(parentItem).toHaveFocus()

      const subtree = getByRole('group')

      // Subtree should now be visible
      expect(subtree).toBeVisible()
    })

    it('moves focus to first child of an expanded item', () => {
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item defaultExpanded>
            Parent
            <TreeView.SubTree>
              <TreeView.Item>Child</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
        </TreeView>
      )

      const parentItem = getByRole('treeitem', {name: 'Parent'})

      // aria-expanded should be true
      expect(parentItem).toHaveAttribute('aria-expanded', 'true')

      // Focus first item
      parentItem.focus()

      // Press →
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowRight'})

      const childItem = getByRole('treeitem', {name: 'Child'})

      // Child item should be focused
      expect(childItem).toHaveFocus()

      // aria-expanded should still be true
      expect(parentItem).toHaveAttribute('aria-expanded', 'true')
    })

    it('does nothing on an end item', () => {
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item defaultExpanded>
            Parent
            <TreeView.SubTree>
              <TreeView.Item>Child 1</TreeView.Item>
              <TreeView.Item>Child 2</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
        </TreeView>
      )

      const parentItem = getByRole('treeitem', {name: 'Parent'})
      const child1 = getByRole('treeitem', {name: 'Child 1'})

      // Focus first item
      parentItem.focus()

      // Press ↓ to move focus to child 1
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})

      // Child 1 should be focused
      expect(child1).toHaveFocus()

      // Press →
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowRight'})

      // Focus should not change
      expect(child1).toHaveFocus()
    })
  })

  describe('Home', () => {
    it('moves focus to first visible item', () => {
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item>
            Parent 1
            <TreeView.SubTree>
              <TreeView.Item>Child 1</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
          <TreeView.Item>
            Parent 2
            <TreeView.SubTree>
              <TreeView.Item>Child 2</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
          <TreeView.Item>
            Parent 3
            <TreeView.SubTree>
              <TreeView.Item>Child 3</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
        </TreeView>
      )

      const parent1 = getByRole('treeitem', {name: 'Parent 1'})
      const parent3 = getByRole('treeitem', {name: 'Parent 2'})

      // Focus first item
      parent1.focus()

      // Press ↓ 2 times to move focus to parent 3
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})

      // Parent 3 should be focused
      expect(parent3).toHaveFocus()

      // Press Home
      fireEvent.keyDown(document.activeElement || document.body, {key: 'Home'})

      // Parent 1 should be focused
      expect(parent1).toHaveFocus()
    })
  })

  describe('End', () => {
    it('moves focus to last visible item', () => {
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item>
            Parent 1
            <TreeView.SubTree>
              <TreeView.Item>Child 1</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
          <TreeView.Item>
            Parent 2
            <TreeView.SubTree>
              <TreeView.Item>Child 2</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
          <TreeView.Item>
            Parent 3
            <TreeView.SubTree>
              <TreeView.Item>Child 3</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
        </TreeView>
      )

      const parent1 = getByRole('treeitem', {name: 'Parent 1'})
      const parent3 = getByRole('treeitem', {name: 'Parent 3'})

      // Focus first item
      parent1.focus()

      // Press End
      fireEvent.keyDown(document.activeElement || document.body, {key: 'End'})

      // Parent 3 should be focused
      expect(parent3).toHaveFocus()

      // Press → to expand parent 3
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowRight'})

      // Press End
      fireEvent.keyDown(document.activeElement || document.body, {key: 'End'})

      const child3 = getByRole('treeitem', {name: 'Child 3'})

      // Child 3 should be focused
      expect(child3).toHaveFocus()
    })
  })

  describe('Enter', () => {
    it('calls onSelect function if provided', () => {
      const onSelect = jest.fn()
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item onSelect={onSelect}>Item</TreeView.Item>
        </TreeView>
      )

      const item = getByRole('treeitem')

      // Focus first item
      item.focus()

      // Press Enter
      fireEvent.keyDown(document.activeElement || document.body, {key: 'Enter'})

      // onSelect should have been called
      expect(onSelect).toHaveBeenCalledTimes(1)
    })

    it('toggles expanded state if no onSelect function is provided', () => {
      const {getByRole, queryByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item>
            Parent
            <TreeView.SubTree>
              <TreeView.Item>Child 1</TreeView.Item>
              <TreeView.Item>Child 2</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
        </TreeView>
      )

      const parent = getByRole('treeitem', {name: 'Parent'})

      // Focus first item
      parent.focus()

      // aria-expanded should be false
      expect(parent).toHaveAttribute('aria-expanded', 'false')

      // Press Enter
      fireEvent.keyDown(document.activeElement || document.body, {key: 'Enter'})

      // aria-expanded should now be true
      expect(parent).toHaveAttribute('aria-expanded', 'true')

      // Subtree should be visible
      expect(queryByRole('group')).toBeVisible()

      // Press Enter
      fireEvent.keyDown(document.activeElement || document.body, {key: 'Enter'})

      // aria-expanded should now be false
      expect(parent).toHaveAttribute('aria-expanded', 'false')

      // Subtree should no longer be visible
      expect(queryByRole('group')).not.toBeInTheDocument()
    })

    it('navigates to href if provided', () => {
      const windowSpy = jest
        .spyOn(window, 'open')
        .mockImplementation(
          (_url?: string | URL | undefined, _target?: string | undefined, _features?: string | undefined) => {
            return null
          }
        )
      const onSelect = jest.fn()
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.LinkItem href="#" onSelect={onSelect}>
            Item
          </TreeView.LinkItem>
        </TreeView>
      )

      const item = getByRole('treeitem')

      // Focus first item
      item.focus()

      // Press Enter
      fireEvent.keyDown(document.activeElement || document.body, {key: 'Enter'})

      // window.open should have been called
      expect(windowSpy).toHaveBeenCalledWith('#', '_self')

      // onSelect should have been called
      expect(onSelect).toHaveBeenCalledTimes(1)
    })
  })

  describe('Typeahead', () => {
    it('moves focus to the next item that matches the typed character', () => {
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item>
            Apple
            <TreeView.SubTree>
              <TreeView.Item>Cantalope</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
          <TreeView.Item>Banana</TreeView.Item>
          <TreeView.Item>Cherry</TreeView.Item>
          <TreeView.Item>Cucumber</TreeView.Item>
        </TreeView>
      )

      const apple = getByRole('treeitem', {name: 'Apple'})
      const cherry = getByRole('treeitem', {name: 'Cherry'})

      // Focus first item
      apple.focus()

      // Press C
      fireEvent.keyDown(document.activeElement || document.body, {key: 'c'})

      // Cherry should be focused
      expect(cherry).toHaveFocus()

      // Notice that the focus is not set to cantalope because
      // it is a child of apple and apple is collapsed.
    })

    it('does nothing if no items match the typed character', () => {
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item>Apple</TreeView.Item>
          <TreeView.Item>Banana</TreeView.Item>
          <TreeView.Item>Cherry</TreeView.Item>
          <TreeView.Item>Durian</TreeView.Item>
        </TreeView>
      )

      const apple = getByRole('treeitem', {name: 'Apple'})

      // Focus first item
      apple.focus()

      // Apple should be focused
      expect(apple).toHaveFocus()

      // Press Z
      fireEvent.keyDown(document.activeElement || document.body, {key: 'z'})

      // Apple should still be focused
      expect(apple).toHaveFocus()
    })

    it('supports multiple typed characters', () => {
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item>Apple</TreeView.Item>
          <TreeView.Item>Banana</TreeView.Item>
          <TreeView.Item>Cherry</TreeView.Item>
          <TreeView.Item>Cantalope 1</TreeView.Item>
          <TreeView.Item>Cantalope 2</TreeView.Item>
        </TreeView>
      )

      const apple = getByRole('treeitem', {name: 'Apple'})
      const cantalope = getByRole('treeitem', {name: 'Cantalope 1'})

      // Focus first item
      apple.focus()

      // Press C + A + N
      fireEvent.keyDown(document.activeElement || document.body, {key: 'c'})
      fireEvent.keyDown(document.activeElement || document.body, {key: 'a'})
      fireEvent.keyDown(document.activeElement || document.body, {key: 'n'})

      // Cantalope should be focused
      expect(cantalope).toHaveFocus()
    })

    it('prioritizes items following the current aria-activedescendant', () => {
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item>Cucumber</TreeView.Item>
          <TreeView.Item current>Cherry</TreeView.Item>
          <TreeView.Item>Cantalope</TreeView.Item>
        </TreeView>
      )

      const cucumber = getByRole('treeitem', {name: 'Cucumber'})
      const cherry = getByRole('treeitem', {name: 'Cherry'})
      const cantalope = getByRole('treeitem', {name: 'Cantalope'})

      // Focus first item
      cucumber.focus()

      // Press ↓ to move focus to cherry
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})

      // Cherry should be focused
      expect(cherry).toHaveFocus()

      // Press C
      fireEvent.keyDown(document.activeElement || document.body, {key: 'c'})

      // Cantalope should be focused
      expect(cantalope).toHaveFocus()
    })

    it('wraps around to the beginning if no items match after the current aria-activedescendant', () => {
      const {getByRole} = renderWithTheme(
        <TreeView aria-label="Test tree">
          <TreeView.Item>Cucumber</TreeView.Item>
          <TreeView.Item>Cherry</TreeView.Item>
          <TreeView.Item current>Cantalope</TreeView.Item>
          <TreeView.Item>Apple</TreeView.Item>
        </TreeView>
      )

      const cantalope = getByRole('treeitem', {name: 'Cantalope'})
      const cucumber = getByRole('treeitem', {name: 'Cucumber'})

      // Focus first item
      cucumber.focus()

      // Press ↓ 2 times to move focus to cantalope
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})
      fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})

      // Cantalope should be focused
      expect(cantalope).toHaveFocus()

      // Press C
      fireEvent.keyDown(document.activeElement || document.body, {key: 'c'})

      // Cucumber should be focused
      expect(cucumber).toHaveFocus()
    })
  })
})

describe('Controlled state', () => {
  it('can be controlled', () => {
    function TestTree() {
      const [expanded, setExpanded] = React.useState(true)
      return (
        <TreeView aria-label="Test tree">
          <TreeView.Item expanded={expanded} onExpandedChange={setExpanded}>
            Parent
            <TreeView.SubTree>
              <TreeView.Item>Child</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
        </TreeView>
      )
    }

    const {getByRole} = renderWithTheme(<TestTree />)

    const parent = getByRole('treeitem', {name: 'Parent'})
    const child = getByRole('treeitem', {name: 'Child'})

    // Parent should be expanded
    expect(parent).toHaveAttribute('aria-expanded', 'true')
    expect(child).toBeVisible()

    // Focus first item
    parent.focus()

    // Press ← to collapse the parent
    fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowLeft'})

    // Parent should be collapsed
    expect(parent).toHaveAttribute('aria-expanded', 'false')
    expect(child).not.toBeVisible()
  })
})

describe('Asyncronous loading', () => {
  it('updates aria live region when loading is done', () => {
    function TestTree() {
      const [state, setState] = React.useState<SubTreeState>('loading')

      return (
        <div>
          {/* Mimic the completion of async loading by clicking the button */}
          <button onClick={() => setState('done')}>Done</button>
          <TreeView aria-label="Test tree">
            <TreeView.Item defaultExpanded>
              Parent
              <TreeView.SubTree state={state}>
                <TreeView.Item>Child</TreeView.Item>
              </TreeView.SubTree>
            </TreeView.Item>
          </TreeView>
        </div>
      )
    }
    const {getByRole} = renderWithTheme(<TestTree />)

    const doneButton = getByRole('button', {name: 'Done'})
    const liveRegion = getByRole('status')

    // Live region should be empty
    expect(liveRegion).toHaveTextContent('')

    // Click done button to mimic the completion of async loading
    fireEvent.click(doneButton)

    // Live region should be updated
    expect(liveRegion).toHaveTextContent('Parent content loaded')
  })

  it('moves focus from loading item to first child', async () => {
    function TestTree() {
      const [state, setState] = React.useState<SubTreeState>('loading')

      React.useEffect(() => {
        const timer = setTimeout(() => setState('done'), 400)
        return () => clearTimeout(timer)
      }, [])

      return (
        <TreeView aria-label="Test tree">
          <TreeView.Item defaultExpanded>
            Parent
            <TreeView.SubTree state={state}>
              <TreeView.Item>Child 1</TreeView.Item>
              <TreeView.Item>Child 2</TreeView.Item>
            </TreeView.SubTree>
          </TreeView.Item>
        </TreeView>
      )
    }

    const {getByRole, findByRole} = renderWithTheme(<TestTree />)

    const parentItem = getByRole('treeitem', {name: 'Parent'})
    const loadingItem = await findByRole('treeitem', {name: 'Loading...'})

    // Focus first item
    parentItem.focus()

    // Press ↓ to move focus to loading item
    fireEvent.keyDown(document.activeElement || document.body, {key: 'ArrowDown'})

    // Loading item should be focused
    expect(loadingItem).toHaveFocus()

    // Wait for async loading to complete
    const firstChild = await findByRole('treeitem', {name: 'Child 1'})

    setTimeout(() => {
      // First child should be focused
      expect(firstChild).toHaveFocus()
    })
  })
})
